/*
   https://www.codeproject.com/Articles/13360/Antialiasing-Wu-Algorithm

   https://en.wikipedia.org/wiki/Xiaolin_Wu%27s_line_algorithm
   https://rosettacode.org/wiki/Xiaolin_Wu%27s_line_algorithm#Java
   https://www.geeksforgeeks.org/anti-aliased-line-xiaolin-wus-algorithm/
   https://unionassets.com/blog/algorithm-brezenhema-and-wu-s-line-299
*/

import java.awt.*;
import javax.swing.*;
import static java.lang.Math.*;

public class XiaolinWuAntialiasing extends JPanel
{
    private final int SIZE = 900;
    private final int NUMBER_OF_LINES = 30;

    public XiaolinWuAntialiasing()
    {
        Dimension dim = new Dimension(SIZE, SIZE);
        setPreferredSize(dim);
        setBackground(Color.white);
    }

    void plot_(Graphics2D g, int x, int y, double c)
    {
        float gray = (float)c;
        g.setColor(new Color(gray, gray, gray));
        g.drawLine(x, y, x, y);
    }

    void plot(Graphics2D g, int x, int y, double c)
    {
        g.setColor(new Color(0f, 0f, 0f, (float)c));
        g.fillOval(x, y, 2, 2);
    }

    double fpart(double x)
    {
        return x - floor(x);
    }

    void drawLine(Graphics2D g, double x0, double y0, double x1, double y1)
    {
        boolean steep = abs(y1 - y0) > abs(x1 - x0);
        if (steep)
        {
           double temp0 = x0;
           x0 = y0;
           y0 = temp0;
           double temp1 = x1;
           x1 = y1;
           y1 = temp1;
        }

        if (x0 > x1)
        {
           double tempX = x0;
           x0 = x1;
           x1 = tempX;
           double tempY = y0;
           y0 = y1;
           y1 = tempY;
        }

        double dx = x1 - x0;
        double dy = y1 - y0;
        double slope = dy / dx;

        // handle first endpoint
        int xend = (int)round(x0);
        double yend = y0 + slope * (xend - x0);
        double xgap = 1.0 - fpart(x0 + 0.5);
        int xpxl1 = xend; // this will be used in the main loop
        int ypxl1 = (int)yend;

        double  fpart = yend - floor(yend);
        double rfpart = 1 - fpart;

        if (steep)
        {
            plot(g, ypxl1,     xpxl1, rfpart * xgap);
            plot(g, ypxl1 + 1, xpxl1,  fpart * xgap);
        }
        else
        {
            plot(g, xpxl1, ypxl1,     rfpart * xgap);
            plot(g, xpxl1, ypxl1 + 1,  fpart * xgap);
        }

        // first y-intersection for the main loop
        double intery = yend + slope;

        // handle second endpoint
        xend = (int)round(x1);
        yend = y1 + slope * (xend - x1);
        xgap = fpart(x1 + 0.5);
        int xpxl2 = xend; // this will be used in the main loop
        int ypxl2 = (int)yend;

         fpart = yend - floor(yend);
        rfpart = 1 - fpart;

        if (steep)
        {
            plot(g, ypxl2,     xpxl2, rfpart * xgap);
            plot(g, ypxl2 + 1, xpxl2,  fpart * xgap);
        }
        else
        {
            plot(g, xpxl2, ypxl2,     rfpart * xgap);
            plot(g, xpxl2, ypxl2 + 1,  fpart * xgap);
        }

        // main loop
        if (steep)
            for (int x = xpxl1 + 1; x <= xpxl2 - 1; x += 1, intery += slope)
            {
                 fpart = intery - floor(intery);
                rfpart = 1 - fpart;
                plot(g, (int)intery,     x, rfpart);
                plot(g, (int)intery + 1, x,  fpart);
            }
        else
            for (int x = xpxl1 + 1; x <= xpxl2 - 1; x += 1, intery += slope)
            {
                 fpart = intery - floor(intery);
                rfpart = 1 - fpart;
                plot(g, x, (int)intery,     rfpart);
                plot(g, x, (int)intery + 1,  fpart);
            }
    }

    @Override
    public void paintComponent(Graphics g)
    {
        super.paintComponent(g);
        Graphics2D g2 = (Graphics2D) g;

      //drawLine(g2, 550, 170, 50, 435);
      for (int i = 0; i <= NUMBER_OF_LINES; i++)
      {
         drawLine(g2, 0, SIZE, SIZE, SIZE - (int)(i * ((double)SIZE/(double)NUMBER_OF_LINES)) );
         drawLine(g2, 0, SIZE, (int)(i * ((double)SIZE/(double)NUMBER_OF_LINES)), 0 );
      }
    }

    public static void main(String[] args)
    {
        SwingUtilities.invokeLater(() ->
        {
            JFrame f = new JFrame();
            f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            f.setTitle("Xiaolin Wu's line algorithm");
            f.setResizable(false);
            f.add(new XiaolinWuAntialiasing(), BorderLayout.CENTER);
            f.pack();
            f.setLocationRelativeTo(null);
            f.setVisible(true);
        });
    }
}
